/*
   This code implements a parallel version of the Gomory-Hu algorithm for
   cut trees using MPI.

   Results appear in the papers:

   Jaime Cohen, Luiz A. Rodrigues, Elias P. Duarte Jr., "Parallel Cut Tree Algorithms", 
   Journal of Parallel and Distributed Computing, ISSN 0743-7315, pp. 1-14, Vol. 109, 2017
   and
   Jaime Cohen, Luiz A. Rodrigues, Elias P. Duarte Jr., "A Parallel
   Implementation of Gomory-Hu's Cut Tree Algorithm," The 24th International Symposium
   on Computer Architecture and High Performance Computing (SBAC-PAD'2012),
   New York, U.S.A., 2012.

   This code is derived from HIPR by IG Systems, Inc.
   HIPR implements a maximum flow - highest level push-relabel algorithm
   http://www.igsystems.com/hipr/
   Commercial use requires a license.
   COPYRIGHT C 1995, 2000 by IG Systems, Inc., igsys@eclipse.net
*/

#include <iostream>
#include <iomanip>
#include <stdio.h>
#include <stdlib.h>
#include <string>
#include "timer.c"                     /* timing routine */
#include <map>
#include <algorithm>
#include <fstream>
#include <sstream>

#include <mpi.h>

#include "hi_pr_contracted.hpp"
#include "hi_pr_input_graph.hpp"
#include "task.hpp"
#include "tree.hpp"

using namespace std;

double THRESHOLD = 0.7;


/***********************************/

string baseNameNoExtension(const string& filename) {
    size_t lastdot = filename.find_last_of(".");
    size_t lastslash = filename.find_last_of("/");
    if (lastslash == std::string::npos) return filename.substr(0, lastdot);;
    return filename.substr(lastslash+1, lastdot-lastslash-1);
}


void update_node_id(int* node_id, int* node_map, int N)
{
  for(int i=N-1; i >= 0; --i)
    node_id[ node_map[i] ] = i;
}

int main (int argc, char *argv[])
{

  double time_init;

  time_init = timer();

  if (argc < 2)
    {
      cerr << argv[0] << " graph.dimacs [run# outpath]" << endl;
      exit(1);
    }

  string fname = argv[1];
  ifstream graph_file(fname.c_str());

#ifdef OUTPUT_STATS
  int run_number=1;
  if (argc >= 3)
    run_number = atoi(argv[2]);
#endif

  string outpath;  // output path (for the stats file)
  if (argc >= 4)
    outpath = string(argv[3])+string("/");
  else outpath = "";

  if (argc  >= 5)
    THRESHOLD = atof(argv[4]);

#ifdef PRINT_STEP_COUNTER
  int step = 0;
#endif

  // common variables
  Edgelist_vector G;   // input graph
  G.read_dimacs(graph_file); // read the input graph
  int N = G.num_nodes();
  int M = G.num_edges();

  Response msg(N); /* contains the cut and its capacity */
  Task task;

  // mpi related variables
  int numprocs, rank, rc;
  // MPI_Status Stat;

  // mpi initialization
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_set_errhandler(MPI_COMM_WORLD, MPI_ERRORS_RETURN);
  //the program will not longer abort on having detected an MPI error, instead the error will be returned and you will have to handle it

  if (numprocs < 2) {
    printf("c Please run with minimum 2 processes.\n");
    fflush(stdout);
    MPI_Finalize();
    exit(1);
  }

  MPI_Barrier(MPI_COMM_WORLD);
  timer1();

  init_task_response_buffers(N);

  // msg.cut = (int *)calloc(N, sizeof(int));
  double sendbuf[3], recvbuf[3]={0.0,0.0,0.0}; //to reduce real, user e sys times

  if (rank==0) {  // master
    GHTree tree(G); // GH tree being constructed
    int failed_tasks = 0;
    long graph_size_sum = 0;
    long number_of_graphs = 0;
    long input_graph_calls = 0;
    long contracted_graph_calls = 0;
    double time_tree_update = 0.0;
    double time_total_to_contract = 0.0;
    double time_flow_contracted = 0.0;
    double time_flow_input_graph = 0.0;
    vector<Slave_state> states(numprocs);

#ifdef DEBUG
    cerr << "c Master, input graph:" << endl;
    cerr << G;
#endif
    int task_number = 0;
    Task task;

#ifdef OUTPUT_STATS
    ofstream stats_file;
    string stats_filename = string(outpath+baseNameNoExtension(string(argv[1]))+".stats.txt");
    stats_file.open(stats_filename.c_str(), fstream::app);
    stats_file << fixed; //  setprecision(9);
    // now the script runOne.sh write the header once
    // stats_file << "num_procs run_number proc s t flow graph_size time_per_cut number_of_cuts balance_factor" << endl;
#endif

    time_init = timer() - time_init;
    // create and send p-1 initial tasks
    for (int p = 1; p < numprocs; p++)
      {
	int s, t;
	GHTreeNode* pivot = tree.pick_pivot_2(); // OK to call pick_pivot_2 (slave will use the input graph anyway)
	// instead of calling tree.build_node_map(pivot) do it directly:
	for (int i = 0; i < N; i++)
	  tree.node_map[i] = i;
	tree.pick_source_and_sink_3(s, t, pivot);
#ifdef DEBUG
	cerr << "c creates task, pivot:" << pivot->id << "  s: " << s << " t: " << t << endl;
#endif
	// Send the task
	task = Task(++task_number, s, t, N, 0, 0); // -1 (no contraction) or 0 (contract)
	sendTask(task, p, N, tree.node_map);
//	states[p].task = task;
	states[p].pivot = pivot->vertices;
      }

    while(1) {
      // receive response
      int from = receiveResponse(msg, N);
      long G_contracted_size;
      if (msg.graph_size == -1) { // mincut found on input graph
	++input_graph_calls;
	G_contracted_size = N;
	time_flow_input_graph += msg.time_per_mincut;
      } else {  // mincut found on a contracted graph
	++contracted_graph_calls;
	G_contracted_size = msg.graph_size;
	time_flow_contracted += msg.time_per_mincut;
      }
      graph_size_sum += G_contracted_size;
#ifdef OUTPUT_STATS
      stats_file << setfill(' ') << setw(4) << numprocs << " " << setw(4) << run_number << " "
		 << setw(4) << from << " " << setw(6) << msg.s+1 << " " << setw(6) << msg.t+1 << " "
		 << setw(12) << setprecision(2) << msg.flow << " " << setprecision(6) << setw(6) << msg.graph_size << " "; // G_contracted_size << " ";
      stats_file << setw(12) << msg.time_per_mincut << " " << setw(6) << msg.num_of_cuts << " " << setw(12)
		 << msg.balance_factor << (tree.failed_task(msg)?" false":" true") << "  " << states[from].pivot.size()  << endl;
#endif
      number_of_graphs++;
      time_total_to_contract += msg.time_to_contract;

#ifdef DEBUG
      cerr << "c Master received response from " <<  from << endl;
      print_reply (msg, N);
#endif

      // Tree update
      double time1 = timer();
      if (!tree.expand_tree(msg)) {
	failed_tasks++; // for statistics
#ifdef DEBUG
	cerr << "Task "<< task.task_number << " failed." << endl;
#endif
      }
      else {
#ifdef DOTFILES
	 tree.to_dot();  // output tree in dot format for visualization
#endif
#ifdef DEBUG
	tree.print(1,1);
#endif
#ifdef PRINT_STEP_COUNTER
	printf("Step %d\n",++step);
#endif
      }
      time_tree_update += timer() - time1;

      if (tree.size() == N)
	break;

      // Make a new task
      int s, t, refine;
      GHTreeNode* pivot = tree.pick_pivot_3(states[from].pivot, refine);
      tree.pick_source_and_sink_3(s, t, pivot);

      // decide whether the contracted graph should be constructed or not
      int contracted_size = pivot->num_vertices + pivot->degree;
#ifdef DEBUG
      cerr << "c creates task, pivot:" << pivot->id << "  s: " << s << " t: " << t << endl;
      cerr << "c contracted_size = " << contracted_size << endl;
#endif
      if (contracted_size < (int)(THRESHOLD*N))
	{
	  // slave should contract the graph of this task
	  int partition_size = tree.build_node_map(pivot);
	  // Send the task
	  task = Task(++task_number, s, t, pivot->num_vertices, partition_size, refine);
	  sendTask(task, from, N, tree.node_map);
	} else
	{ // slave should use the input graph (partition_size = -1 flag)
	  task = Task(++task_number, s, t, pivot->num_vertices, -1, refine);
	  sendTask(task, from, N, tree.node_map);
	}
//      states[from].task = task;
      states[from].pivot = pivot->vertices;
    }

    timer2();
    sendbuf[0] = getRealTime();
    sendbuf[1] = getUserTime();
    sendbuf[2] = getSystemTime();

    // receive all pending replies (Jaime 3/2015)
    // receive from all slaves except 'from'
    for (int p = 1; p < numprocs-1; ++p)
      receiveResponse(msg, N);

    //sending source=-1 to all slaves to make then stop
    for (int p = 1; p < numprocs; p++)
      {
	Task task = Task(++task_number, -1, -1, 0, 0, 0);
	sendTask(task, p, N, tree.node_map);
#ifdef DEBUG
	printf("c Master sent task s= %d t=%d to Slave %d\n", 0, p, p);
#endif
      }

    // OUTPUT THE ANSWER
    tree.print(0,1);

    MPI_Reduce( &sendbuf, &recvbuf, 3, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    cout << "c timer: real: "<< sendbuf[0] << " user: " << recvbuf[1] << " sys: " << recvbuf[2] << endl;
    cout << "c success: " << N-1 << " unsuccess: " << failed_tasks << endl;
    cout << "c mean graph size: " << graph_size_sum / number_of_graphs << endl;
    cout << "c input graph calls: " << input_graph_calls << endl;
    cout << "c contracted graph calls: " << contracted_graph_calls << endl;
    cout << "c number of max flow calls: " << failed_tasks + N - 1 << endl;
    cout << "c time initialization: " << time_init << endl;
    cout << "c time tree update: " << time_tree_update << endl;
    cout << "c time flow contracted: " << time_flow_contracted << endl;
    cout << "c time flow input graph: " << time_flow_input_graph << endl;
    cout << "c time contract: " << time_total_to_contract << endl;
    cout << "c threshold: " << THRESHOLD << endl;
    MPI_Finalize();
#ifdef OUTPUT_STATS
    stats_file.close();
#endif
    release_task_response_buffers();
    return (0);

  } else { //SLAVES

    // allocate space for the contracted graphs
    // assume that the contracted graphs are smaller than the input graph
    Edgelist G_contracted(N, M) ; // contracted graph
    Edgelist old_G_contracted(N, M);
    Edgelist *G_c = &G_contracted;
    Edgelist *old_G_c = &old_G_contracted;
    bool has_Gc = false;
    hi_pr_init(N, M);

    // read the input graph once and allocate data structure for
    // the max. flows on the input graph
    parse_input_graph(fname);

    int* node_map = new int[N];
    int* node_map_Gc = new int[N];
    int* node_id = new int[N];
    vector<bool> cut(N);
    int source, sink;

#ifdef DEBUGFILE
    ofstream debug_file;
    stringstream tmp_out;
    tmp_out << rank;
    string debug_filename = string("out")+tmp_out.str();
    debug_file.open(debug_filename);
#endif

#ifdef DEBUG
    printf("c      Slave %d running\n",rank);
#endif

    while (1) {
      rc = receiveTask(task, N, node_map);
#ifdef DEBUG
      cerr << "c      Slave << " << rank << " received task s = " << task.s
	   << " t = " << task.t << endl;
      print_task(task, node_map, N);
#endif

      if (rc != MPI_SUCCESS || task.s < 0) { //receive finish - stop slave
         timer2();
         sendbuf[0] = getRealTime();
         sendbuf[1] = getUserTime();
         sendbuf[2] = getSystemTime();

         // cout << "c Thread " << rank << " finalized. Timer: real: " << sendbuf[0] <<  " user: "
	 //     << sendbuf[1] <<  " sys:" << sendbuf[2] << endl;
         MPI_Reduce( &sendbuf, &recvbuf, 3, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
         MPI_Finalize();
	 release_task_response_buffers();
         return (0);
      }
      // heuristic: the source is the vertex with minimum weighted degree among s and t

      if (task.part_len >= 0)  // use the contracted graph
	{
	  double start_time = timer();
	  // BUILD CONTRACTED GRAPH
	  msg.graph_size =  task.pivot_size + task.part_len;
	  source = node_map[task.s];
	  sink = node_map[task.t];
	  bool updated_gc = true;
	  if (has_Gc && task.refine) {
	    // if (msg.graph_size <= THRESHOLD2 * G_c->num_nodes()) {
	      std::swap(old_G_c, G_c);
	      build_contracted_graph_from_contracted(*old_G_c, *G_c, node_map, msg.graph_size, node_id);
	      // } else {
	      //source = node_map_Gc[task.s];
	      //sink = node_map_Gc[task.t];
	      //updated_gc = false;
	      //std::swap(node_map, node_map_Gc);
	      //}
	  } else { // build contracted from input graph
	      build_contracted_graph(G, *G_c, node_map, msg.graph_size);
	      has_Gc = true;
	  }
	  msg.time_to_contract = timer() - start_time;

#ifdef DEBUGFILE
	  debug_file << "---------------" << endl;
	  // print_node_map(node_map, N, debug_file);
	  debug_file << task.s+1 << "/" << task.t+1 << endl;
	  debug_file << "s = " << node_map[task.s] << " t = " << node_map[task.t]  << endl;
	  debug_file << "G_contracted:" << endl;
	  debug_file << *G_c;
	  debug_file.flush();
#endif

	  bool inverted = false;
#ifndef NODEGREE
      if (G_c->degrees[source] > G_c->degrees[sink])
	{
	  std::swap(source,sink);
	  inverted = true;
	}
#endif

      G_c->set_source(source);
      G_c->set_sink(sink);
      if (updated_gc) {
    	  update_node_id(node_id, node_map, N);
    	  parse_contracted(*G_c);
      }
      // CALL HI_PR with balanced cut heuristic or not
#ifdef BALANCED
	  // search for balanced cuts
	  msg.flow = hi_pr_contracted(source, sink, cut, true, msg.time_per_mincut,
			   msg.num_of_cuts, msg.balance_factor);
#else
	  msg.flow = hi_pr_contracted(source, sink, cut, false, msg.time_per_mincut,
			   msg.num_of_cuts, msg.balance_factor);
#endif
	  if (!inverted) {
	    for (int i = 0; i < N; i++)
	      msg.cut[i] = cut[node_map[i]];
	  } else {
	    for (int i = 0; i < N; i++)
	      msg.cut[i] = 1 - cut[node_map[i]];
	  }
	  std::swap(node_map, node_map_Gc);
	} else // use the input graph
	{
	  int source = task.s;
	  int sink = task.t;
	  int inverted = false;
#ifndef NODEGREE
	  if (G.degrees[source] > G.degrees[sink])
	    {
	      std:: swap(source,sink);
	      inverted = true;
	    }
#endif
	  // use input graph (non-contracted)
#ifdef BALANCED
	  msg.flow = hi_pr_on_input_graph(source, sink, cut, true, msg.time_per_mincut,
					  msg.num_of_cuts, msg.balance_factor);
#else
	  msg.flow = hi_pr_on_input_graph(source, sink, cut, false, msg.time_per_mincut,
					  msg.num_of_cuts, msg.balance_factor);
#endif
	  if (!inverted) {
	    for (int i = 0; i < N; i++)
	      msg.cut[i] = cut[i];
 	  } else {
	    for (int i = 0; i < N; i++)
	      msg.cut[i] = 1-cut[i];
	  }
	  msg.graph_size = -1;
	  has_Gc = false;
	}
#ifdef DEBUGFILE
      debug_file << "hi_pr cut: " << task.s+1 << " " << task.t+1 << " --> ";
      for (int i = 0; i < N; i++)
	if (msg.cut[i])
	  debug_file << i+1 << " ";
      debug_file << endl;
#endif

      //send results to master
      msg.task_number = task.task_number;
      msg.s = task.s;
      msg.t = task.t;

      rc = sendResponse(msg, N);
    }//end while
    delete [] node_map;
    delete [] node_id;
  } // end SLAVE

//  MPI_Abort(MPI_COMM_WORLD, 0);
//  MPI_Finalize();

  release_task_response_buffers();
  return 0;

}
